from typing import Union

def get_core_utils() -> dict[str, callable]:
    """ Returns the core utility functions. """
    return {
        "get_val_from_dict": _get_val_from_dict,
        "concatenate_strings": _concatenate_strings,
        "_mean": _mean,
    }

def _get_val_from_dict(key: str, dictionary: dict[str, any]) -> any:
    """ Gets a value from a dictionary. """
    return dictionary[key]

def _concatenate_strings(lst: list[str]) -> str:
    """ Concatenates a list of strings. """
    return "".join(lst)

def _mean(lst: list[Union[int, float]]) -> float:
    """ Calculates the mean of a list of numbers. """
    return sum(lst) / len(lst)


import os

from groq import Groq

from easy_fnc.utils import get_template_path
from easy_fnc.models.model import EasyFNCModel

class GroqModel(EasyFNCModel):
    """A class to interact with the Groq API and chat with the model"""
    def __init__(
            self, 
            functions: list[dict[str, str]],
            model_name: str = os.environ.get("GROQ_MODEL"),
            template_path: str = get_template_path(),
            template_type: str = "toml"
        ):
        super().__init__(functions, template_path, template_type)
        self.client = Groq(api_key=os.environ.get("GROQ_API_KEY"))
        self.model_name = model_name
        self.messages = [{"role": "system", "content": self.generate_system_prompt().replace("<|user_query|>", "")}]

        if os.environ.get("GROQ_MODEL") is None:
            print("No GROQ_MODEL environment variable found. Using: llama3-8b-8192.")
            self.model_name = "llama3-8b-8192"
        else: 
            self.model_name = os.environ.get("GROQ_MODEL")

        if os.environ.get("GROQ_API_KEY") is None:
            print("No GROQ_API_KEY environment variable found. Please set it to your Groq API key.")

    def generate(self, user_message: str) -> str:
        """Chat with the model and return the response"""
        # Create a message object for the user input
        self.messages.append({"role": "user", "content": self.format_user_input(user_message)})

        # Get the chat completion from the model
        return self._get_chat_completion(self.messages)

    def _get_chat_completion(self, messages) -> str:
        """Get the chat completion from the model"""
        model_response = self.client.chat.completions.create(messages=messages, model=self.model_name).choices[0].message.content

        # Add the model response to the messages
        self.messages.append({"role": "assistant", "content": model_response})

        return model_response


import inspect
from ast import literal_eval

from easy_fnc.functions import get_user_defined_functions
from easy_fnc.core_utils import get_core_utils
from easy_fnc.schemas import FunctionCall, ModelResponse
from easy_fnc.utils import extract_thoughts_and_function_calls

class FunctionCallingEngine:
    """
    Function calling engine for EasyFNC.
    """
    def __init__(
            self,
            extraction_function: callable = extract_thoughts_and_function_calls,
        ):
        # Initialize the functions dictionary
        core_utils = get_core_utils()
        self.functions = {**core_utils}
        self.outputs = {}
        self.extraction_function = extraction_function

    def add_user_functions(self, file_path: str):
        """
        Add user-defined functions from the specified file path.
        """
        user_functions = get_user_defined_functions(file_path)
        self.functions.update(user_functions)

    def parse_model_response(self, raw_response: str) -> ModelResponse:
        """
        Parse the model response and return the ModelResponse object.
        """
        return ModelResponse.from_raw_response(raw_response, self.extraction_function)
    
    def call_functions(self, function_calls: list[FunctionCall]) -> dict:
        """
        Call the functions from the given input.
        """
        def map_previous_outputs(outputs_dict: dict, inputs_dict: dict) -> dict:
            """Map the previous outputs to the input."""
            for key, value in inputs_dict.items():
                if isinstance(value, str):
                    if value in outputs_dict:
                        inputs_dict[key] = outputs_dict[value]
                if isinstance(value, dict):
                    inputs_dict[key] = map_previous_outputs(value)

            return inputs_dict
        
        for function_call in function_calls:
            function_name = function_call.name
            function_input = function_call.kwargs

            # Check if the input is an output from a previous function
            function_input = map_previous_outputs(self.outputs, function_input)

            if function_call.returns:
                # Switch case for length of returns
                match len(function_call.returns):
                    case 0:
                        # Call the function with no returns
                        output = self.functions[function_name]()
                        self.outputs[function_call.returns[0]] = output
                    case 1:
                        # Call the function with one return
                        output = self.functions[function_name](**function_input)
                        self.outputs[function_call.returns[0]] = output
                    case _:
                        # Call the function with one or more returns
                        outputs_tuple = self.functions[function_name](**function_input) 
                        for i in range(len(function_call.returns)):
                            self.outputs[function_call.returns[i]] = outputs_tuple[i]

        return self.outputs if self.outputs else {}

def create_functions_metadata(functions: dict) -> list[dict]:
        """Creates the functions metadata for the prompt. """
        def format_type(p_type: str) -> str:
            """Format the type of the parameter."""
            # If p_type begins with "<class", then it is a class type
            if p_type.startswith("<class"):
                # Get the class name from the type
                p_type = p_type.split("'")[1]
            
            return p_type
            
        functions_metadata = []
        i = 0
        for name, function in functions.items():
            i += 1
            descriptions = function.__doc__.split("\n")
            functions_metadata.append({
                "name": name,
                "description": descriptions[0],
                "parameters": {
                    "properties": [ # Get the parameters for the function
                        {   
                            "name": param_name,
                            "type": format_type(str(param_type)),
                        }
                        # Remove the return type from the parameters
                        for param_name, param_type in function.__annotations__.items() if param_name != "return"
                    ],
                    
                    "required": [param_name for param_name in function.__annotations__ if param_name != "return"],
                } if function.__annotations__ else {},
                "returns": [
                    {
                        "name": name + "_output",
                        "type": {param_name: format_type(str(param_type)) for param_name, param_type in function.__annotations__.items() if param_name == "return"}["return"]
                    }
                ]
            })

        return functions_metadata


from pydantic import BaseModel, Field

import json

from easy_fnc.utils import extract_thoughts_and_function_calls

class FunctionCall(BaseModel):
    """
    Pydantic model for the function call.
    """
    name: str = Field(..., description="The name of the function to be called")
    kwargs: dict = Field({}, description="The keyword arguments for the function to be called")
    returns: list[str] = Field([], description="The return values of the function that is called")

    @classmethod
    def from_dict(cls, data: dict):
        """
        Create a FunctionCall object from a dictionary.
        """
        return cls(**data)

class ModelResponse(BaseModel):
    """
    Pydantic model for the model response.
    """
    thoughts: str = Field(..., description="The thoughts of the model")
    function_calls: list[FunctionCall] = Field([], description="The function calls made by the model")

    @classmethod
    def from_raw_response(
        cls, 
        raw_response: str,
        extraction_function: callable = extract_thoughts_and_function_calls
    ):
        """
        Create a ModelResponse object from the raw response.

        Args:
        - raw_response (str): 
            The raw response from the model.
        - extraction_function (callable): 
            The function to extract the thoughts and function calls from the raw 
            response. Should return a tuple of thoughts and function calls string.
        """
        # Extract the thoughts and function calls
        thoughts, function_calls_str = extraction_function(raw_response)

        if len(thoughts) < 1 and len(function_calls_str) < 1:
            raise ValueError("No thoughts or function calls found in the raw response. Raw response: \n" + raw_response)
    
        # Parse the function calls
        try:
            function_calls_loaded = json.loads(function_calls_str)
        except json.JSONDecodeError:
            raise ValueError("Error parsing function calls from the raw response. Raw response: \n" + raw_response)
        
        function_calls = [FunctionCall.from_dict(data) for data in function_calls_loaded]
        
        return cls(thoughts=thoughts, function_calls=function_calls)
    
    def __str__(self):
        """
        Return a string representation of the ModelResponse object.
        """
        output = "Thoughts:\n"
        output += self.thoughts.strip() + "\n\n"
        output += "Function Calls:\n"
        for function_call in self.function_calls:
            output += f"- {function_call.name}:\n"
            output += f"  - kwargs: {function_call.kwargs}\n"
            output += f"  - returns: {function_call.returns}\n"
        return output


import random
import importlib
import ast

def get_user_defined_functions(filename: str) -> dict[str, callable]:
    """Retrieves the user defined functions from a file"""
    # Parse the file
    try:
        with open(filename, 'r') as file:
            tree = ast.parse(file.read())
    except FileNotFoundError:
        raise FileNotFoundError(f"File {filename} not found")

    # Get the user defined functions
    functions: dict[str, callable] = {}
    for node in ast.walk(tree):
        if isinstance(node, ast.FunctionDef):
            # Import function by name
            module = importlib.import_module(filename.split(".")[0].replace("/", "."))
            function = getattr(module, node.name)

            # Append the function to the list (exclude the get_user_defined_functions function)
            if node.name != "get_user_defined_functions":
                functions[node.name] = function

    return functions

# Example functions
def get_weather_forecast(location: str) -> dict[str, str]:
    """Retrieves the weather forecast for a given location"""
    forecasts = ["sunny", "cloudy", "rainy", "snowy", "windy"]
    return {
        "location": location,
        "forecast": random.choice(forecasts),
        "temperature": f"{random.randint(-10, 40)}°C"
    }

def get_random_city() -> str:
    """Retrieves a random city from a list of cities"""
    cities = ["Groningen", "Enschede", "Amsterdam", "Istanbul", "Baghdad", "Rio de Janeiro", "Tokyo", "Kampala"]
    return random.choice(cities)

def get_random_number(low: int, high: int) -> int:
    """Retrieves a random number between low and high"""
    return random.randint(low, high)



import ollama

import json

from easy_fnc.models.model import EasyFNCModel
from easy_fnc.utils import get_template_path

class OllamaModel(EasyFNCModel):
    """
    A class to generate responses using the Ollama model.
    """
    def __init__(
            self, 
            model_name: str, 
            functions: list[dict[str, str]],
            template_path: str = get_template_path(),
            template_type: str = "toml"
        ) -> None:
        super().__init__(functions, template_path, template_type)
        self.model_name = model_name
        self.messages = [{"role": "system", "content": self.generate_system_prompt().replace("<|user_query|>", "")}]

    def format_output(self, output: any, original_prompt: str) -> str:
        """ Format the output. """
        prompt = self.template["function_response_prompt"]["beginning"] + original_prompt
        prompt += self.template["function_response_prompt"]["middle"]
        prompt += json.dumps(output, indent=4)
        prompt += self.template["function_response_prompt"]["end"]

        return prompt
    
    def generate(
            self, 
            user_input: str
            ) -> str:
        """
        Generate a response based on the user input.
        """
        # Get the model response and extract the content
        self.messages.append({"role": "user", "content": self.format_user_input(user_input)})
        model_response = ollama.chat(
            model=self.model_name,
            messages=self.messages
        )

        content = model_response["message"]["content"].strip().replace("\'", "\"")

        return content


from abc import ABC, abstractmethod
import json

from easy_fnc.utils import load_template, get_template_path

class EasyFNCModel(ABC):
    """
    Abstract class for EasyFNC models.
    """
    def __init__(
            self, 
            functions: list[dict[str, str]],
            template_path: str = get_template_path(),
            template_type: str = "toml"
        ) -> None:
        self.functions = functions
        self.template = load_template(file_path=template_path, file_type=template_type)

    def generate_system_prompt(self) -> str:
        """ Generate the system prompt. """
        prompt_beginning = self.template["function_call_prompt"]["beginning"]
        system_prompt_end = self.template["function_call_prompt"]["system_prompt_end"]
        return prompt_beginning + json.dumps(self.functions, indent=4) + system_prompt_end
    
    def format_user_input(self, user_input: str) -> str:
        """ Format the user input. """
        return "<|user_query|>" + user_input + "<|end_user_query|>"

    @abstractmethod
    def generate(self, user_input: str) -> str:
        """
        Generate a response based on the user input.
        """
        pass
    



import json
import tomllib
import re
import os

# Define constants
TEMPLATE_FILE_TYPES = ["json", "toml"]

def get_template_path():
    """
    Return the path of the base.toml file.
    """
    # Get the current directory
    current_dir = os.path.dirname(os.path.abspath(__file__))

    # Return the path of the base.toml file
    return os.path.join(current_dir, "models", "templates", "base.toml")

def load_template(
        file_path: str = get_template_path(),
        file_type: str = "toml"
    ) -> dict:
    """
    Load a template from a JSON file and return it as a dictionary.
    """
    if file_type not in TEMPLATE_FILE_TYPES:
        raise ValueError(f"Invalid file type. Expected one of {TEMPLATE_FILE_TYPES}")
    if file_type == "json":
        return load_json(file_path)
    elif file_type == "toml":
        return load_toml(file_path)
    

def load_json(file_path: str) -> dict:
    """
    Load a JSON file and return it as a dictionary.
    """
    try:
        with open(file_path, "r") as file:
            data = json.load(file)
    except FileNotFoundError:
        raise FileNotFoundError(f"File {file_path} not found")
    return data

def load_toml(file_path: str) -> dict:
    """
    Load a TOML file and return it as a dictionary.
    """
    try:
        with open(file_path, "rb") as file:
            data = tomllib.load(file)
    except FileNotFoundError:
        raise FileNotFoundError(f"File {file_path} not found")
    except tomllib.TOMLDecodeError:
        raise tomllib.TOMLDecodeError(f"Error decoding TOML file {file_path}")
    return data

def extract_thoughts_and_function_calls(raw_response: str) -> tuple[str, str]:
    """
    Extract the thoughts and function calls from the raw model response.
    """
    pattern = r'<\|thoughts\|>(.*?)<\|end_thoughts\|>\s*<\|function_calls\|>(.*?)<\|end_function_calls\|>'
    match = re.search(pattern, raw_response, re.DOTALL)
    
    if match:
        thoughts = match.group(1).strip()
        function_calls = match.group(2).strip()
        return thoughts, function_calls
    else:
        return "", ""


from setuptools import setup, find_packages

# Declare the constants
PYPI_PACKAGES = ["setuptools", "wheel", "twine"]

def read_readme():
    """Read the README.md file and return its contents."""
    with open("README.md", "r", encoding="utf-8") as fh:
        return fh.read()
    
def read_requirements() -> list[str]:
    """Read the requirements.txt file and return its contents."""
    def validate_line(line: str) -> bool:
        """Check if the line is valid."""
        line = line.strip()
        if line.startswith("#") or line == "\n" or line in PYPI_PACKAGES or line == "":
            return False
        return True
    
    with open("requirements.txt", "r", encoding="utf-8") as fh:
        return [line.strip() for line in fh if validate_line(line)]


def main():
    setup(
        name='easy_fnc',
        version='0.2.1',
        description='This package hopes to provide a modular and highly extendable interface to interact with LLMs via (multiple) function calling, easily.',
        long_description=read_readme(),
        long_description_content_type="text/markdown",
        author='Atakan Tekparmak',
        author_email='atakantekerparmak@gmail.com',
        url="https://github.com/AtakanTekparmak/easy_fnc",
        packages=find_packages(),
        install_requires=read_requirements(),
        extras_require={
            "groq": ["groq"],
            "ollama": ["ollama"],
        }
    )

if __name__ == "__main__":
    main()